@@ -39,7 +39,7 @@ typedef enum
     OPDATA_IMPORT, OPDATA_GENKEY, OPDATA_KEYLIST, OPDATA_EDIT,
     OPDATA_VERIFY, OPDATA_TRUSTLIST, OPDATA_ASSUAN, OPDATA_VFS_MOUNT,
     OPDATA_PASSWD, OPDATA_EXPORT, OPDATA_KEYSIGN, OPDATA_TOFU_POLICY,
-    OPDATA_QUERY_SWDB
+    OPDATA_QUERY_SWDB, OPDATA_SETEXPIRE, OPDATA_REVSIG
   } ctx_op_data_id_t;
 
 
@@ -118,6 +118,25 @@ struct gpgme_context
    * flag is cleared with each operation.  */
   unsigned int redraw_suggested : 1;
 
+  /* True if the option --include-key-block shall be passed to gpg.  */
+  unsigned int include_key_block : 1;
+
+  /* True if the option --auto-key-import shall be passed to gpg.  */
+  unsigned int auto_key_import : 1;
+
+  /* True if the option --auto-key-retrieve shall be passed to gpg.  */
+  unsigned int auto_key_retrieve : 1;
+
+  /* Do not use the symmtric encryption passphrase cache.  */
+  unsigned int no_symkey_cache : 1;
+
+  /* Pass --ignore-mdc-error to gpg.  Note that this flag is reset
+   * after the operation.  */
+  unsigned int ignore_mdc_error : 1;
+
+  /* Pass --expert to gpg edit key. */
+  unsigned int extended_edit : 1;
+
   /* Flags for keylist mode.  */
   gpgme_keylist_mode_t keylist_mode;
 
@@ -142,10 +161,28 @@ struct gpgme_context
   /* The gpg specific override session key or NULL. */
   char *override_session_key;
 
+  /* The optional request origin.  */
+  char *request_origin;
+
+  /* The optional auto key locate options.  */
+  char *auto_key_locate;
+
   /* The locale for the pinentry.  */
   char *lc_ctype;
   char *lc_messages;
 
+  /* The optional trust-model override.  */
+  char *trust_model;
+
+  /* The optional expiration date of a certification.  */
+  char *cert_expire;
+
+  /* The optional key origin.  */
+  char *key_origin;
+
+  /* The optional import filter.  */
+  char *import_filter;
+
   /* The operation data hooked into the context.  */
   ctx_op_data_t op_data;
 
