@@ -16,8 +16,9 @@
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, see <http://www.gnu.org/licenses/>.
- **/
+ * License along with this program; if not, see <https://gnu.org/licenses/>.
+ * SPDX-License-Identifier: LGPL-2.1-or-later
+ */
 
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -65,13 +66,22 @@
 #include "sys-util.h"
 
 
-#ifndef HAVE_W32CE_SYSTEM
 #define HAVE_ALLOW_SET_FOREGROUND_WINDOW 1
-#endif
 #ifndef F_OK
 # define F_OK 0
 #endif
 
+/* The Registry key used by GNUPG.  */
+#ifdef _WIN64
+# define GNUPG_REGKEY_2  "Software\\Wow6432Node\\GNU\\GnuPG"
+#else
+# define GNUPG_REGKEY_2  "Software\\GNU\\GnuPG"
+#endif
+#ifdef _WIN64
+# define GNUPG_REGKEY_3  "Software\\Wow6432Node\\GnuPG"
+#else
+# define GNUPG_REGKEY_3  "Software\\GnuPG"
+#endif
 
 DEFINE_STATIC_LOCK (get_path_lock);
 
@@ -85,9 +95,10 @@ static HMODULE my_hmodule;
    binaries.  The are set only once by gpgme_set_global_flag.  */
 static char *default_gpg_name;
 static char *default_gpgconf_name;
-
-
-#ifdef HAVE_ALLOW_SET_FOREGROUND_WINDOW
+/* If this variable is not NULL the value is assumed to be the
+   installation directory.  The variable may only be set once by
+   gpgme_set_global_flag and accessed by _gpgme_get_inst_dir.  */
+static char *override_inst_dir;
 
 #define RTLD_LAZY 0
 
@@ -95,6 +106,8 @@ static GPG_ERR_INLINE void *
 dlopen (const char * name, int flag)
 {
   void * hd = LoadLibrary (name);
+
+  (void)flag;
   return hd;
 }
 
@@ -121,7 +134,6 @@ dlclose (void * hd)
     }
   return -1;
 }
-#endif /* HAVE_ALLOW_SET_FOREGROUND_WINDOW */
 
 
 /* Return a malloced string encoded in UTF-8 from the wide char input
@@ -156,6 +168,48 @@ wchar_to_utf8 (const wchar_t *string)
 }
 
 
+/* Return a malloced wide char string from an UTF-8 encoded input
+   string STRING.  Caller must free this value. On failure returns
+   NULL; caller may use GetLastError to get the actual error number.
+   Calling this function with STRING set to NULL is not defined. */
+static wchar_t *
+utf8_to_wchar (const char *string)
+{
+  int n;
+  wchar_t *result;
+
+
+  n = MultiByteToWideChar (CP_UTF8, 0, string, -1, NULL, 0);
+  if (n < 0)
+    return NULL;
+
+  result = (wchar_t *) malloc ((n+1) * sizeof *result);
+  if (!result)
+    return NULL;
+
+  n = MultiByteToWideChar (CP_UTF8, 0, string, -1, result, n);
+  if (n < 0)
+    {
+      free (result);
+      return NULL;
+    }
+  return result;
+}
+
+
+/* Same as utf8_to_wchar but calling it with NULL returns
+   NULL.  So a return value of NULL only indicates failure
+   if STRING is not set to NULL. */
+static wchar_t *
+utf8_to_wchar0 (const char *string)
+{
+  if (!string)
+    return NULL;
+
+  return utf8_to_wchar (string);
+}
+
+
 /* Replace all forward slashes by backslashes.  */
 static void
 replace_slashes (char *string)
@@ -208,25 +262,67 @@ _gpgme_allow_set_foreground_window (pid_t pid)
 
   if (!pid || pid == (pid_t)(-1))
     {
-      TRACE1 (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", NULL,
 	      "no action for pid %d", (int)pid);
     }
   else if (func)
     {
       int rc = func (pid);
-      TRACE2 (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", NULL,
 	      "called for pid %d; result=%d", (int)pid, rc);
 
     }
   else
     {
-      TRACE0 (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", 0,
+      TRACE (DEBUG_ENGINE, "gpgme:AllowSetForegroundWindow", NULL,
 	      "function not available");
     }
 #endif /* HAVE_ALLOW_SET_FOREGROUND_WINDOW */
 }
 
 
+/* Wrapper around CancelSynchronousIo which is only available since
+ * Vista.  */
+void
+_gpgme_w32_cancel_synchronous_io (HANDLE thread)
+{
+  static int initialized;
+  static BOOL (WINAPI * func)(DWORD);
+  void *handle;
+
+  if (!initialized)
+    {
+      /* Available since Vista; thus we dynload it.  */
+      initialized = 1;
+      handle = dlopen ("kernel32.dll", RTLD_LAZY);
+      if (handle)
+        {
+          func = dlsym (handle, "CancelSynchronousIo");
+          if (!func)
+            {
+              dlclose (handle);
+              handle = NULL;
+            }
+        }
+    }
+
+  if (func)
+    {
+      if (!func ((DWORD)thread) && GetLastError() != ERROR_NOT_FOUND)
+        {
+          TRACE (DEBUG_ENGINE, "gpgme:CancelSynchronousIo", NULL,
+                 "called for thread %p: ec=%u",
+                 thread, (unsigned int)GetLastError ());
+        }
+    }
+  else
+    {
+      TRACE (DEBUG_ENGINE, "gpgme:CancelSynchronousIo", NULL,
+	      "function not available");
+    }
+}
+
+
 /* Return a string from the W32 Registry or NULL in case of error.
    Caller must release the return value.  A NULL for root is an alias
    for HKEY_CURRENT_USER, HKEY_LOCAL_MACHINE in turn. */
@@ -287,53 +383,6 @@ read_w32_registry_string (const char *root, const char *dir, const char *name)
     }
   result[nbytes] = 0; /* Make sure it is really a string.  */
 
-#ifndef HAVE_W32CE_SYSTEM
-  /* Windows CE does not have an environment.  */
-  if (type == REG_EXPAND_SZ && strchr (result, '%'))
-    {
-      char *tmp;
-
-      n1 += 1000;
-      tmp = malloc (n1 + 1);
-      if (!tmp)
-        goto leave;
-      nbytes = ExpandEnvironmentStrings (result, tmp, n1);
-      if (nbytes && nbytes > n1)
-        {
-          free (tmp);
-          n1 = nbytes;
-          tmp = malloc (n1 + 1);
-          if (!tmp)
-            goto leave;
-          nbytes = ExpandEnvironmentStrings (result, tmp, n1);
-          if (nbytes && nbytes > n1) {
-            free (tmp); /* Oops - truncated, better don't expand at all. */
-            goto leave;
-          }
-          tmp[nbytes] = 0;
-          free (result);
-          result = tmp;
-        }
-      else if (nbytes)  /* Okay, reduce the length. */
-        {
-          tmp[nbytes] = 0;
-          free (result);
-          result = malloc (strlen (tmp)+1);
-          if (!result)
-            result = tmp;
-          else
-            {
-              strcpy (result, tmp);
-              free (tmp);
-            }
-        }
-      else  /* Error - don't expand. */
-        {
-          free (tmp);
-        }
-    }
-#endif
-
  leave:
   RegCloseKey (key_handle);
   return result;
@@ -347,6 +396,9 @@ _gpgme_get_inst_dir (void)
 {
   static char *inst_dir;
 
+  if (override_inst_dir)
+    return override_inst_dir;
+
   LOCK (get_path_lock);
   if (!inst_dir)
     {
@@ -382,12 +434,11 @@ find_program_in_dir (const char *dir, const char *name)
 {
   char *result;
 
-  result = malloc (strlen (dir) + 1 + strlen (name) + 1);
+  result = _gpgme_strconcat (dir, "\\", name, NULL);
   if (!result)
     return NULL;
 
-  strcpy (stpcpy (stpcpy (result, dir), "\\"), name);
-  if (access (result, F_OK))
+  if (_gpgme_access (result, F_OK))
     {
       free (result);
       return NULL;
@@ -400,21 +451,35 @@ find_program_in_dir (const char *dir, const char *name)
 static char *
 find_program_at_standard_place (const char *name)
 {
-  char path[MAX_PATH];
+  wchar_t path[MAX_PATH];
   char *result = NULL;
 
-  /* See http://wiki.tcl.tk/17492 for details on compatibility.  */
-  if (SHGetSpecialFolderPathA (NULL, path, CSIDL_PROGRAM_FILES, 0))
+  /* See https://wiki.tcl-lang.org/page/Getting+Windows+%22special+folders%22+with+Ffidl for details on compatibility.
+
+     We First try the generic place and then fallback to the x86
+     (i.e. 32 bit) place.  This will prefer a 64 bit of the program
+     over a 32 bit version on 64 bit Windows if installed.  */
+  if (SHGetSpecialFolderPathW (NULL, path, CSIDL_PROGRAM_FILES, 0))
     {
-      result = malloc (strlen (path) + 1 + strlen (name) + 1);
-      if (result)
+      char *utf8_path = wchar_to_utf8 (path);
+      result = _gpgme_strconcat (utf8_path, "\\", name, NULL);
+      free (utf8_path);
+      if (result && _gpgme_access (result, F_OK))
         {
-          strcpy (stpcpy (stpcpy (result, path), "\\"), name);
-          if (access (result, F_OK))
-            {
-              free (result);
-              result = NULL;
-            }
+          free (result);
+          result = NULL;
+        }
+    }
+  if (!result
+      && SHGetSpecialFolderPathW (NULL, path, CSIDL_PROGRAM_FILESX86, 0))
+    {
+      char *utf8_path = wchar_to_utf8 (path);
+      result = _gpgme_strconcat (utf8_path, "\\", name, NULL);
+      free (utf8_path);
+      if (result && _gpgme_access (result, F_OK))
+        {
+          free (result);
+          result = NULL;
         }
     }
   return result;
@@ -428,12 +493,9 @@ _gpgme_set_default_gpg_name (const char *name)
 {
   if (!default_gpg_name)
     {
-      default_gpg_name = malloc (strlen (name) + 5);
+      default_gpg_name = _gpgme_strconcat (name, ".exe", NULL);
       if (default_gpg_name)
-        {
-          strcpy (stpcpy (default_gpg_name, name), ".exe");
-          replace_slashes (default_gpg_name);
-        }
+        replace_slashes (default_gpg_name);
     }
   return !default_gpg_name;
 }
@@ -445,17 +507,35 @@ _gpgme_set_default_gpgconf_name (const char *name)
 {
   if (!default_gpgconf_name)
     {
-      default_gpgconf_name = malloc (strlen (name) + 5);
+      default_gpgconf_name = _gpgme_strconcat (name, ".exe", NULL);
       if (default_gpgconf_name)
-        {
-          strcpy (stpcpy (default_gpgconf_name, name), ".exe");
-          replace_slashes (default_gpgconf_name);
-        }
+        replace_slashes (default_gpgconf_name);
     }
   return !default_gpgconf_name;
 }
 
 
+/* Set the override installation directory.  This function may only be
+   called by gpgme_set_global_flag.  Returns 0 on success.  */
+int
+_gpgme_set_override_inst_dir (const char *dir)
+{
+  if (!override_inst_dir)
+    {
+      override_inst_dir = strdup (dir);
+      if (override_inst_dir)
+        {
+          replace_slashes (override_inst_dir);
+          /* Remove a trailing slash.  */
+          if (*override_inst_dir
+              && override_inst_dir[strlen (override_inst_dir)-1] == '\\')
+            override_inst_dir[strlen (override_inst_dir)-1] = 0;
+        }
+    }
+  return !override_inst_dir;
+}
+
+
 /* Return the full file name of the GPG binary.  This function is used
    iff gpgconf was not found and thus it can be assumed that gpg2 is
    not installed.  This function is only called by get_gpgconf_item
@@ -481,7 +561,7 @@ _gpgme_get_gpg_path (void)
       char *dir;
 
       dir = read_w32_registry_string ("HKEY_LOCAL_MACHINE",
-                                      "Software\\GNU\\GnuPG",
+                                      GNUPG_REGKEY_2,
                                       "Install Directory");
       if (dir)
         {
@@ -499,7 +579,8 @@ _gpgme_get_gpg_path (void)
 
   /* 4. Print a debug message if not found.  */
   if (!gpg)
-    _gpgme_debug (DEBUG_ENGINE, "_gpgme_get_gpg_path: '%s' not found", name);
+    _gpgme_debug (NULL, DEBUG_ENGINE, -1, NULL, NULL, NULL,
+                  "_gpgme_get_gpg_path: '%s' not found", name);
 
   return gpg;
 }
@@ -530,15 +611,27 @@ _gpgme_get_gpgconf_path (void)
       gpgconf = find_program_at_standard_place (name2);
     }
 
-  /* 3. Try to find gpgconf.exe using that ancient registry key.  This
-        should eventually be removed.  */
+  /* 3. Try to find gpgconf.exe using the Windows registry. */
   if (!gpgconf)
     {
       char *dir;
 
-      dir = read_w32_registry_string ("HKEY_LOCAL_MACHINE",
-                                      "Software\\GNU\\GnuPG",
+      dir = read_w32_registry_string (NULL,
+                                      GNUPG_REGKEY_2,
                                       "Install Directory");
+      if (!dir)
+        {
+          char *tmp = read_w32_registry_string (NULL,
+                                                GNUPG_REGKEY_3,
+                                                "Install Directory");
+          if (tmp)
+            {
+              dir = _gpgme_strconcat (tmp, "\\bin", NULL);
+              free (tmp);
+              if (!dir)
+                return NULL;
+            }
+        }
       if (dir)
         {
           gpgconf = find_program_in_dir (dir, name);
@@ -552,9 +645,18 @@ _gpgme_get_gpgconf_path (void)
       gpgconf = find_program_at_standard_place ("GNU\\GnuPG\\gpgconf.exe");
     }
 
+  /* 5. Try to find gpgconf.exe relative to us.  */
+  if (!gpgconf && inst_dir)
+    {
+      char *dir = _gpgme_strconcat (inst_dir, "\\..\\..\\GnuPG\\bin", NULL);
+      gpgconf = find_program_in_dir (dir, name);
+      free (dir);
+    }
+
   /* 5. Print a debug message if not found.  */
   if (!gpgconf)
-    _gpgme_debug (DEBUG_ENGINE, "_gpgme_get_gpgconf_path: '%s' not found",name);
+    _gpgme_debug (NULL, DEBUG_ENGINE, -1, NULL, NULL, NULL,
+                  "_gpgme_get_gpgconf_path: '%s' not found",name);
 
   return gpgconf;
 }
@@ -589,15 +691,8 @@ _gpgme_get_conf_int (const char *key, int *value)
   return 1;
 }
 
-
-#ifdef HAVE_W32CE_SYSTEM
-int
-_gpgme_mkstemp (int *fd, char **name)
-{
-  return -1;
-}
-#else
 
+
 /* mkstemp extracted from libc/sysdeps/posix/tempname.c.  Copyright
    (C) 1991-1999, 2000, 2001, 2006 Free Software Foundation, Inc.
 
@@ -614,7 +709,7 @@ static const char letters[] =
    does not exist at the time of the call to mkstemp.  TMPL is
    overwritten with the result.  */
 static int
-mkstemp (char *tmpl)
+my_mkstemp (char *tmpl)
 {
   int len;
   char *XXXXXX;
@@ -718,12 +813,11 @@ _gpgme_mkstemp (int *fd, char **name)
 	}
     }
 
-  tmpname = malloc (strlen (tmp) + 13 + 1);
+  tmpname = _gpgme_strconcat (tmp, "\\gpgme-XXXXXX", NULL);
   if (!tmpname)
     return -1;
-  strcpy (stpcpy (tmpname, tmp), "\\gpgme-XXXXXX");
-  *fd = mkstemp (tmpname);
-  if (fd < 0)
+  *fd = my_mkstemp (tmpname);
+  if (*fd < 0)
     {
       free (tmpname);
       return -1;
@@ -732,30 +826,77 @@ _gpgme_mkstemp (int *fd, char **name)
   *name = tmpname;
   return 0;
 }
-#endif
 
 
-
-#ifdef HAVE_W32CE_SYSTEM
-/* Return a malloced string with the replacement value for the
-   GPGME_DEBUG envvar.  Caller must release.  Returns NULL if not
-   set.  */
-char *
-_gpgme_w32ce_get_debug_envvar (void)
+/* Like access but using windows _waccess */
+int
+_gpgme_access (const char *path, int mode)
 {
-  char *tmp;
+  wchar_t *u16 = utf8_to_wchar0 (path);
+  int r = _waccess (u16, mode);
 
-  tmp = read_w32_registry_string (NULL, "\\Software\\GNU\\gpgme", "debug");
-  if (tmp && !*tmp)
-    {
-      free (tmp);
-      tmp = NULL;
-    }
-  return tmp;
+  free(u16);
+  return r;
 }
-#endif /*HAVE_W32CE_SYSTEM*/
 
 
+/* Like CreateProcessA but mapping the arguments to wchar API */
+int
+_gpgme_create_process_utf8 (const char *application_name_utf8,
+                            char *command_line_utf8,
+                            LPSECURITY_ATTRIBUTES lpProcessAttributes,
+                            LPSECURITY_ATTRIBUTES lpThreadAttributes,
+                            BOOL bInheritHandles,
+                            DWORD dwCreationFlags,
+                            void *lpEnvironment,
+                            char *working_directory_utf8,
+                            LPSTARTUPINFOA si,
+                            LPPROCESS_INFORMATION lpProcessInformation)
+{
+  BOOL ret;
+  wchar_t *application_name = utf8_to_wchar0 (application_name_utf8);
+  wchar_t *command_line = utf8_to_wchar0 (command_line_utf8);
+  wchar_t *working_directory = utf8_to_wchar0 (working_directory_utf8);
+
+  STARTUPINFOW siw;
+  memset (&siw, 0, sizeof siw);
+  if (si)
+    {
+      siw.cb = sizeof (siw);
+      siw.dwFlags = si->dwFlags;
+      siw.wShowWindow = si->wShowWindow;
+      siw.hStdInput = si->hStdInput;
+      siw.hStdOutput = si->hStdOutput;
+      siw.hStdError = si->hStdError;
+      siw.dwX = si->dwX;
+      siw.dwY = si->dwY;
+      siw.dwXSize = si->dwXSize;
+      siw.dwYSize = si->dwYSize;
+      siw.dwXCountChars = si->dwXCountChars;
+      siw.dwYCountChars = si->dwYCountChars;
+      siw.dwFillAttribute = si->dwFillAttribute;
+      siw.lpDesktop = utf8_to_wchar0 (si->lpDesktop);
+      siw.lpTitle = utf8_to_wchar0 (si->lpTitle);
+    }
+
+  ret = CreateProcessW (application_name,
+                        command_line,
+                        lpProcessAttributes,
+                        lpThreadAttributes,
+                        bInheritHandles,
+                        dwCreationFlags,
+                        lpEnvironment,
+                        working_directory,
+                        si ? &siw : NULL,
+                        lpProcessInformation);
+  free (siw.lpTitle);
+  free (siw.lpDesktop);
+  free (application_name);
+  free (command_line);
+  free (working_directory);
+  return ret;
+}
+
 /* Entry point called by the DLL loader.  */
 #ifdef DLL_EXPORT
 int WINAPI
